@@ -50,6 +50,7 @@
 
 #include "llvm/ADT/StringSwitch.h"
 #include "llvm/Bitcode/BitcodeWriter.h"
+#include "llvm/Demangle/Demangle.h"
 #include "llvm/IR/IRBuilder.h"
 #include "llvm/IR/IntrinsicInst.h"
 #include "llvm/IR/Metadata.h"
@@ -89,14 +90,22 @@ void removeFnAttr(CallInst *Call, Attribute::AttrKind Attr) {
   Call->removeFnAttr(Attr);
 }
 
-Value *removeCast(Value *V) {
-  auto Cast = dyn_cast<ConstantExpr>(V);
-  if (Cast && Cast->isCast()) {
-    return removeCast(Cast->getOperand(0));
+Value *extendVector(Value *V, FixedVectorType *NewType,
+                    IRBuilderBase &Builder) {
+  unsigned OldSize = cast<FixedVectorType>(V->getType())->getNumElements();
+  unsigned NewSize = NewType->getNumElements();
+  assert(OldSize < NewSize);
+  std::vector<Constant *> Components;
+  IntegerType *Int32Ty = Builder.getInt32Ty();
+  for (unsigned I = 0; I < NewSize; I++) {
+    if (I < OldSize)
+      Components.push_back(ConstantInt::get(Int32Ty, I));
+    else
+      Components.push_back(PoisonValue::get(Int32Ty));
   }
-  if (auto Cast = dyn_cast<CastInst>(V))
-    return removeCast(Cast->getOperand(0));
-  return V;
+
+  return Builder.CreateShuffleVector(V, PoisonValue::get(V->getType()),
+                                     ConstantVector::get(Components), "vecext");
 }
 
 void saveLLVMModule(Module *M, const std::string &OutputFile) {
@@ -216,10 +225,14 @@ std::string mapSPIRVTypeToOCLType(SPIRVType *Ty, bool Signed) {
 
 PointerType *getOrCreateOpaquePtrType(Module *M, const std::string &Name,
                                       unsigned AddrSpace) {
+  return PointerType::get(getOrCreateOpaqueStructType(M, Name), AddrSpace);
+}
+
+StructType *getOrCreateOpaqueStructType(Module *M, StringRef Name) {
   auto OpaqueType = StructType::getTypeByName(M->getContext(), Name);
   if (!OpaqueType)
     OpaqueType = StructType::create(M->getContext(), Name);
-  return PointerType::get(OpaqueType, AddrSpace);
+  return OpaqueType;
 }
 
 PointerType *getSamplerType(Module *M) {
@@ -227,9 +240,9 @@ PointerType *getSamplerType(Module *M) {
                                   SPIRAS_Constant);
 }
 
-PointerType *getPipeStorageType(Module *M) {
-  return getOrCreateOpaquePtrType(
-      M, getSPIRVTypeName(kSPIRVTypeName::PipeStorage), SPIRAS_Constant);
+Type *getSamplerStructType(Module *M) {
+  return getOrCreateOpaqueStructType(M,
+                                     getSPIRVTypeName(kSPIRVTypeName::Sampler));
 }
 
 PointerType *getSPIRVOpaquePtrType(Module *M, Op OC) {
@@ -246,67 +259,39 @@ void getFunctionTypeParameterTypes(llvm::FunctionType *FT,
 
 bool isVoidFuncTy(FunctionType *FT) { return FT->getReturnType()->isVoidTy(); }
 
-bool isPointerToOpaqueStructType(llvm::Type *Ty) {
-  if (auto PT = dyn_cast<PointerType>(Ty))
-    if (auto *ST = dyn_cast<StructType>(PT->getPointerElementType()))
-      if (ST->isOpaque())
+bool isOCLImageStructType(llvm::Type *Ty, StringRef *Name) {
+  if (auto *ST = dyn_cast_or_null<StructType>(Ty))
+    if (ST->isOpaque()) {
+      auto FullName = ST->getName();
+      if (FullName.find(kSPR2TypeName::ImagePrefix) == 0) {
+        if (Name)
+          *Name = FullName.drop_front(strlen(kSPR2TypeName::OCLPrefix));
         return true;
-  return false;
-}
-
-bool isPointerToOpaqueStructType(llvm::Type *Ty, const std::string &Name) {
-  if (auto PT = dyn_cast<PointerType>(Ty))
-    if (auto *ST = dyn_cast<StructType>(PT->getPointerElementType()))
-      if (ST->isOpaque() && ST->getName() == Name)
-        return true;
-  return false;
-}
-
-bool isSPIRVSamplerType(llvm::Type *Ty) {
-  if (auto *PT = dyn_cast<PointerType>(Ty))
-    if (auto *ST = dyn_cast<StructType>(PT->getPointerElementType()))
-      if (ST->isOpaque()) {
-        auto Name = ST->getName();
-        if (Name.startswith(std::string(kSPIRVTypeName::PrefixAndDelim) +
-                            kSPIRVTypeName::Sampler)) {
-          return true;
-        }
-      }
-  return false;
-}
-
-bool isOCLImageType(llvm::Type *Ty, StringRef *Name) {
-  if (auto PT = dyn_cast<PointerType>(Ty))
-    if (auto *ST = dyn_cast<StructType>(PT->getPointerElementType()))
-      if (ST->isOpaque()) {
-        auto FullName = ST->getName();
-        if (FullName.find(kSPR2TypeName::ImagePrefix) == 0) {
-          if (Name)
-            *Name = FullName.drop_front(strlen(kSPR2TypeName::OCLPrefix));
-          return true;
-        }
       }
+    }
   return false;
 }
 
 /// \param BaseTyName is the type Name as in spirv.BaseTyName.Postfixes
 /// \param Postfix contains postfixes extracted from the SPIR-V image
 ///   type Name as spirv.BaseTyName.Postfixes.
-bool isSPIRVType(llvm::Type *Ty, StringRef BaseTyName, StringRef *Postfix) {
-  if (auto PT = dyn_cast<PointerType>(Ty))
-    if (auto *ST = dyn_cast<StructType>(PT->getPointerElementType()))
-      if (ST->isOpaque()) {
-        auto FullName = ST->getName();
-        std::string N =
-            std::string(kSPIRVTypeName::PrefixAndDelim) + BaseTyName.str();
-        if (FullName != N)
-          N = N + kSPIRVTypeName::Delimiter;
-        if (FullName.startswith(N)) {
-          if (Postfix)
-            *Postfix = FullName.drop_front(N.size());
-          return true;
-        }
-      }
+bool isSPIRVStructType(llvm::Type *Ty, StringRef BaseTyName,
+                       StringRef *Postfix) {
+  auto *ST = dyn_cast<StructType>(Ty);
+  if (!ST)
+    return false;
+  if (ST->isOpaque()) {
+    auto FullName = ST->getName();
+    std::string N =
+        std::string(kSPIRVTypeName::PrefixAndDelim) + BaseTyName.str();
+    if (FullName != N)
+      N = N + kSPIRVTypeName::Delimiter;
+    if (FullName.startswith(N)) {
+      if (Postfix)
+        *Postfix = FullName.drop_front(N.size());
+      return true;
+    }
+  }
   return false;
 }
 
@@ -325,6 +310,21 @@ bool isSYCLHalfType(llvm::Type *Ty) {
   return false;
 }
 
+bool isSYCLBfloat16Type(llvm::Type *Ty) {
+  if (auto *ST = dyn_cast<StructType>(Ty)) {
+    if (!ST->hasName())
+      return false;
+    StringRef Name = ST->getName();
+    Name.consume_front("class.");
+    if ((Name.startswith("cl::sycl::") ||
+         Name.startswith("__sycl_internal::")) &&
+        Name.endswith("::bfloat16")) {
+      return true;
+    }
+  }
+  return false;
+}
+
 Function *getOrCreateFunction(Module *M, Type *RetTy, ArrayRef<Type *> ArgTypes,
                               StringRef Name, BuiltinFuncMangleInfo *Mangle,
                               AttributeList *Attrs, bool TakeName) {
@@ -628,24 +628,6 @@ bool containsUnsignedAtomicType(StringRef Name) {
       Name[Loc + strlen(kMangledName::AtomicPrefixIncoming)]);
 }
 
-bool isFunctionPointerType(Type *T) {
-  if (isa<PointerType>(T) && isa<FunctionType>(T->getPointerElementType())) {
-    return true;
-  }
-  return false;
-}
-
-bool hasFunctionPointerArg(Function *F, Function::arg_iterator &AI) {
-  AI = F->arg_begin();
-  for (auto AE = F->arg_end(); AI != AE; ++AI) {
-    LLVM_DEBUG(dbgs() << "[hasFuncPointerArg] " << *AI << '\n');
-    if (isFunctionPointerType(AI->getType())) {
-      return true;
-    }
-  }
-  return false;
-}
-
 Constant *castToVoidFuncPtr(Function *F) {
   auto T = getVoidFuncPtrType(F->getParent());
   return ConstantExpr::getBitCast(F, T);
@@ -661,6 +643,131 @@ bool hasArrayArg(Function *F) {
   return false;
 }
 
+/// Convert a struct name from the name given to it in Itanium name mangling to
+/// the name given to it as an LLVM opaque struct.
+static std::string demangleBuiltinOpenCLTypeName(StringRef MangledStructName) {
+  assert(MangledStructName.startswith("ocl_") &&
+         "Not a valid builtin OpenCL mangled name");
+  // Bare structure type that starts with ocl_ is a builtin opencl type.
+  // See clang/lib/CodeGen/CGOpenCLRuntime for how these map to LLVM types
+  // and clang/lib/AST/ItaniumMangle for how they are mangled.
+  // In general, ocl_<foo> is mapped to pointer-to-%opencl.<foo>, but
+  // there is some variance around whether or not _t is included in the
+  // mangled name.
+  std::string LlvmStructName = StringSwitch<StringRef>(MangledStructName)
+                                   .Case("ocl_sampler", "opencl.sampler_t")
+                                   .Case("ocl_event", "opencl.event_t")
+                                   .Case("ocl_clkevent", "opencl.clk_event_t")
+                                   .Case("ocl_queue", "opencl.queue_t")
+                                   .Case("ocl_reserveid", "opencl.reserve_id_t")
+                                   .Default("")
+                                   .str();
+  if (LlvmStructName.empty()) {
+    LlvmStructName = "opencl.";
+    LlvmStructName += MangledStructName.substr(4); // Strip off ocl_
+    if (!MangledStructName.endswith("_t"))
+      LlvmStructName += "_t";
+  }
+  return LlvmStructName;
+}
+
+void getParameterTypes(Function *F, SmallVectorImpl<StructType *> &ArgTys) {
+  // If there's no mangled name, we can't do anything. Also, if there's no
+  // parameters, do nothing.
+  StringRef Name = F->getName();
+  if (!Name.startswith("_Z") || F->arg_empty())
+    return;
+
+  // Start by filling in a skeleton of information we can get from the LLVM type
+  // itself.
+  ArgTys.clear();
+  auto *FT = F->getFunctionType();
+  ArgTys.reserve(FT->getNumParams());
+  bool HasSret = false;
+  for (Argument &Arg : F->args()) {
+    if (!Arg.getType()->isPointerTy())
+      ArgTys.push_back(nullptr);
+    else if (Type *Ty = Arg.getParamStructRetType()) {
+      assert(!HasSret && &Arg == F->getArg(0) &&
+             "sret parameter should only appear on the first argument");
+      HasSret = true;
+      ArgTys.push_back(dyn_cast<StructType>(Ty));
+    } else {
+      ArgTys.push_back(nullptr);
+    }
+  }
+
+  Module *M = F->getParent();
+
+  // Skip the first argument if it's an sret parameter--this would be an
+  // implicit parameter not recognized as part of the function parameters.
+  auto *ArgIter = ArgTys.begin();
+  if (HasSret)
+    ++ArgIter;
+
+  // Demangle the function arguments. If we get an input name of
+  // "_Z12write_imagei20ocl_image1d_array_woDv2_iiDv4_i", then we expect
+  // that Demangler.getFunctionParameters will return
+  // "(ocl_image1d_array_wo, int __vector(2), int, int __vector(4))" (in other
+  // words, the stuff between the parentheses if you ran C++ filt, including
+  // the parentheses itself).
+  ItaniumPartialDemangler Demangler;
+  std::string MangledName = F->getName().str();
+  if (Demangler.partialDemangle(MangledName.c_str()))
+    return;
+  char *Buf = nullptr;
+  size_t BufLen = 0;
+  Buf = Demangler.getFunctionParameters(Buf, &BufLen);
+  StringRef Args(Buf, BufLen);
+  // Strip parentheses from the result.
+  Args = Args.slice(1, Args.size() - 2);
+
+  if (Args.find_first_of("<(") == StringRef::npos) {
+    // Go through the function arguments using type names where possible.
+    SmallVector<StringRef, 8> ArgParams;
+    Args.split(ArgParams, ", ");
+
+    // Sanity check that the name mangling matches up to the expected number of
+    // arguments.
+    if (ArgParams.size() > (size_t)(ArgTys.end() - ArgIter)) {
+      LLVM_DEBUG(dbgs() << "[getParameterTypes] function " << F->getName()
+                        << " appears to have " << ArgParams.size()
+                        << " arguments but has " << (ArgTys.end() - ArgIter)
+                        << "\n");
+      free(Buf);
+      return;
+    }
+
+    for (StringRef Arg : ArgParams) {
+      StructType *Pointee = nullptr;
+      if (Arg.endswith("*") && !Arg.endswith("**")) {
+        // Strip off address space and other qualifiers.
+        StringRef MangledStructName = Arg.split(' ').first;
+
+        if (MangledStructName.consume_front("__spirv_")) {
+          // This is a pointer to a SPIR-V OpType* opaque struct. In general,
+          // convert __spirv_<Type>[__Suffix] to %spirv.Type[._Suffix]
+          auto NameSuffixPair = MangledStructName.split('_');
+          std::string StructName = "spirv.";
+          StructName += NameSuffixPair.first;
+          if (!NameSuffixPair.second.empty()) {
+            StructName += ".";
+            StructName += NameSuffixPair.second;
+          }
+          Pointee = getOrCreateOpaqueStructType(M, StructName);
+        } else if (MangledStructName.startswith("opencl.")) {
+          Pointee = getOrCreateOpaqueStructType(M, MangledStructName);
+        }
+      } else if (!Arg.contains(' ') && Arg.startswith("ocl_")) {
+        std::string StructName = demangleBuiltinOpenCLTypeName(Arg);
+        Pointee = getOrCreateOpaqueStructType(M, StructName);
+      }
+      *ArgIter++ = Pointee;
+    }
+  }
+  free(Buf);
+}
+
 CallInst *mutateCallInst(
     Module *M, CallInst *CI,
     std::function<std::string(CallInst *, std::vector<Value *> &)> ArgMutate,
@@ -677,6 +784,13 @@ CallInst *mutateCallInst(
   auto NewCI = addCallInst(M, NewName, CI->getType(), Args, Attrs, CI, Mangle,
                            InstName, TakeFuncName);
   NewCI->setDebugLoc(CI->getDebugLoc());
+  NewCI->copyMetadata(*CI);
+  NewCI->setAttributes(CI->getAttributes());
+  NewCI->setTailCall(CI->isTailCall());
+  if (CI->hasFnAttr("fpbuiltin-max-error")) {
+    auto Attr = CI->getFnAttr("fpbuiltin-max-error");
+    NewCI->addFnAttr(Attr);
+  }
   LLVM_DEBUG(dbgs() << " => " << *NewCI << '\n');
   CI->replaceAllUsesWith(NewCI);
   CI->eraseFromParent();
@@ -917,9 +1031,9 @@ Metadata *getMDOperandOrNull(MDNode *N, unsigned I) {
   return N->getOperand(I);
 }
 
-std::string getMDOperandAsString(MDNode *N, unsigned I) {
+StringRef getMDOperandAsString(MDNode *N, unsigned I) {
   if (auto *Str = dyn_cast_or_null<MDString>(getMDOperandOrNull(N, I)))
-    return Str->getString().str();
+    return Str->getString();
   return "";
 }
 
@@ -945,7 +1059,7 @@ std::set<std::string> getNamedMDAsStringSet(Module *M,
     if (!MD || MD->getNumOperands() == 0)
       continue;
     for (unsigned J = 0, N = MD->getNumOperands(); J != N; ++J)
-      StrSet.insert(getMDOperandAsString(MD, J));
+      StrSet.insert(getMDOperandAsString(MD, J).str());
   }
 
   return StrSet;
@@ -1098,7 +1212,7 @@ static SPIR::RefParamType transTypeDesc(Type *Ty,
       return SPIR::RefParamType(new SPIR::PrimitiveType(
           Signed ? SPIR::PRIMITIVE_LONG : SPIR::PRIMITIVE_ULONG));
     default:
-      llvm_unreachable("invliad int size");
+      return SPIR::RefParamType(new SPIR::PrimitiveType(SPIR::PRIMITIVE_INT));
     }
   }
   if (Ty->isVoidTy())
@@ -1239,8 +1353,12 @@ Value *getScalarOrArrayConstantInt(Instruction *Pos, Type *T, unsigned Len,
     assert(Len == 1 && "Invalid length");
     return ConstantInt::get(IT, V, IsSigned);
   }
-  if (auto PT = dyn_cast<PointerType>(T)) {
-    auto ET = PT->getPointerElementType();
+  if (isa<PointerType>(T)) {
+    unsigned PointerSize =
+        Pos->getModule()->getDataLayout().getPointerTypeSizeInBits(T);
+    auto *ET = Type::getIntNTy(T->getContext(), PointerSize);
+    assert(cast<PointerType>(T)->isOpaqueOrPointeeTypeMatches(ET) &&
+           "Pointer-to-non-size_t arguments are not valid for this call");
     auto AT = ArrayType::get(ET, Len);
     std::vector<Constant *> EV(Len, ConstantInt::get(ET, V, IsSigned));
     auto CA = ConstantArray::get(AT, EV);
@@ -1292,9 +1410,17 @@ bool isSPIRVConstantName(StringRef TyName) {
 
 Type *getSPIRVTypeByChangeBaseTypeName(Module *M, Type *T, StringRef OldName,
                                        StringRef NewName) {
+  return PointerType::get(
+      getSPIRVStructTypeByChangeBaseTypeName(M, T, OldName, NewName),
+      SPIRAS_Global);
+}
+
+Type *getSPIRVStructTypeByChangeBaseTypeName(Module *M, Type *T,
+                                             StringRef OldName,
+                                             StringRef NewName) {
   StringRef Postfixes;
-  if (isSPIRVType(T, OldName, &Postfixes))
-    return getOrCreateOpaquePtrType(M, getSPIRVTypeName(NewName, Postfixes));
+  if (isSPIRVStructType(T, OldName, &Postfixes))
+    return getOrCreateOpaqueStructType(M, getSPIRVTypeName(NewName, Postfixes));
   LLVM_DEBUG(dbgs() << " Invalid SPIR-V type " << *T << '\n');
   llvm_unreachable("Invalid SPIR-V type");
   return nullptr;
@@ -1325,6 +1451,11 @@ std::string getSPIRVImageSampledTypeName(SPIRVType *Ty) {
       else
         return kSPIRVImageSampledTypeName::UInt;
     }
+    if (Ty->getIntegerBitWidth() == 64) {
+      if (static_cast<SPIRVTypeInt *>(Ty)->isSigned())
+        return kSPIRVImageSampledTypeName::Long;
+      return kSPIRVImageSampledTypeName::ULong;
+    }
     break;
   case OpTypeFloat:
     switch (Ty->getFloatBitWidth()) {
@@ -1356,6 +1487,9 @@ Type *getLLVMTypeForSPIRVImageSampledTypePostfix(StringRef Postfix,
   if (Postfix == kSPIRVImageSampledTypeName::Int ||
       Postfix == kSPIRVImageSampledTypeName::UInt)
     return Type::getInt32Ty(Ctx);
+  if (Postfix == kSPIRVImageSampledTypeName::Long ||
+      Postfix == kSPIRVImageSampledTypeName::ULong)
+    return Type::getInt64Ty(Ctx);
   llvm_unreachable("Invalid sampled type postfix");
   return nullptr;
 }
@@ -1378,6 +1512,18 @@ std::string getImageBaseTypeName(StringRef Name) {
   return ImageTyName;
 }
 
+size_t getImageOperandsIndex(Op OpCode) {
+  switch (OpCode) {
+  case OpImageRead:
+  case OpImageSampleExplicitLod:
+    return 2;
+  case OpImageWrite:
+    return 3;
+  default:
+    return ~0U;
+  }
+}
+
 std::string mapOCLTypeNameToSPIRV(StringRef Name, StringRef Acc) {
   std::string BaseTy;
   std::string Postfixes;
@@ -1558,13 +1704,16 @@ StringRef getAccessQualifierFullName(StringRef TyName) {
 }
 
 /// Translates OpenCL image type names to SPIR-V.
-Type *getSPIRVImageTypeFromOCL(Module *M, Type *ImageTy) {
-  assert(isOCLImageType(ImageTy) && "Unsupported type");
-  auto ImageTypeName = ImageTy->getPointerElementType()->getStructName();
-  StringRef Acc = kAccessQualName::ReadOnly;
-  if (hasAccessQualifiedName(ImageTypeName))
-    Acc = getAccessQualifierFullName(ImageTypeName);
-  return getOrCreateOpaquePtrType(M, mapOCLTypeNameToSPIRV(ImageTypeName, Acc));
+Type *adaptSPIRVImageType(Module *M, Type *PointeeType) {
+  if (isOCLImageStructType(PointeeType)) {
+    auto ImageTypeName = PointeeType->getStructName();
+    StringRef Acc = kAccessQualName::ReadOnly;
+    if (hasAccessQualifiedName(ImageTypeName))
+      Acc = getAccessQualifierFullName(ImageTypeName);
+    return getOrCreateOpaqueStructType(
+        M, mapOCLTypeNameToSPIRV(ImageTypeName, Acc));
+  }
+  return PointeeType;
 }
 
 llvm::PointerType *getOCLClkEventType(Module *M) {
@@ -1679,7 +1828,8 @@ bool checkTypeForSPIRVExtendedInstLowering(IntrinsicInst *II, SPIRVModule *BM) {
       Ty = VecTy->getElementType();
     }
     if ((!Ty->isFloatTy() && !Ty->isDoubleTy() && !Ty->isHalfTy()) ||
-        ((NumElems > 4) && (NumElems != 8) && (NumElems != 16))) {
+        (!BM->hasCapability(CapabilityVectorAnyINTEL) &&
+         ((NumElems > 4) && (NumElems != 8) && (NumElems != 16)))) {
       BM->SPIRVCK(
           false, InvalidFunctionCall, II->getCalledOperand()->getName().str());
       return false;
@@ -1694,7 +1844,8 @@ bool checkTypeForSPIRVExtendedInstLowering(IntrinsicInst *II, SPIRVModule *BM) {
       Ty = VecTy->getElementType();
     }
     if ((!Ty->isIntegerTy()) ||
-        ((NumElems > 4) && (NumElems != 8) && (NumElems != 16))) {
+        (!BM->hasCapability(CapabilityVectorAnyINTEL) &&
+         ((NumElems > 4) && (NumElems != 8) && (NumElems != 16)))) {
       BM->SPIRVCK(
           false, InvalidFunctionCall, II->getCalledOperand()->getName().str());
     }
@@ -1706,14 +1857,15 @@ bool checkTypeForSPIRVExtendedInstLowering(IntrinsicInst *II, SPIRVModule *BM) {
   return true;
 }
 
-void setAttrByCalledFunc(CallInst *Call) {
+CallInst *setAttrByCalledFunc(CallInst *Call) {
   Function *F = Call->getCalledFunction();
   assert(F);
   if (F->isIntrinsic()) {
-    return;
+    return Call;
   }
   Call->setCallingConv(F->getCallingConv());
   Call->setAttributes(F->getAttributes());
+  return Call;
 }
 
 bool isSPIRVBuiltinVariable(GlobalVariable *GV,
@@ -1763,6 +1915,94 @@ bool isSPIRVBuiltinVariable(GlobalVariable *GV,
 // %4 = call spir_func i64 @_Z13get_global_idj(i32 2) #1
 // %5 = insertelement <3 x i64> %3, i64 %4, i32 2
 // %6 = extractelement <3 x i64> %5, i32 0
+
+/// Recursively look through the uses of a global variable, including casts or
+/// gep offsets, to find all loads of the variable. Gep offsets that are non-0
+/// are accumulated in the AccumulatedOffset parameter, which will eventually be
+/// used to figure out which index of a variable is being used.
+static void replaceUsesOfBuiltinVar(Value *V, const APInt &AccumulatedOffset,
+                                    Function *ReplacementFunc,
+                                    GlobalVariable *GV) {
+  const DataLayout &DL = ReplacementFunc->getParent()->getDataLayout();
+  SmallVector<Instruction *, 4> InstsToRemove;
+  for (User *U : V->users()) {
+    if (auto *Cast = dyn_cast<CastInst>(U)) {
+      replaceUsesOfBuiltinVar(Cast, AccumulatedOffset, ReplacementFunc, GV);
+      InstsToRemove.push_back(Cast);
+    } else if (auto *GEP = dyn_cast<GetElementPtrInst>(U)) {
+      APInt NewOffset = AccumulatedOffset.sextOrTrunc(
+          DL.getIndexSizeInBits(GEP->getPointerAddressSpace()));
+      if (!GEP->accumulateConstantOffset(DL, NewOffset))
+        llvm_unreachable("Illegal GEP of a SPIR-V builtin variable");
+      replaceUsesOfBuiltinVar(GEP, NewOffset, ReplacementFunc, GV);
+      InstsToRemove.push_back(GEP);
+    } else if (auto *Load = dyn_cast<LoadInst>(U)) {
+      // Figure out which index the accumulated offset corresponds to. If we
+      // have a weird offset (e.g., trying to load byte 7), bail out.
+      Type *ScalarTy = ReplacementFunc->getReturnType();
+      APInt Index;
+      uint64_t Remainder;
+      APInt::udivrem(AccumulatedOffset, ScalarTy->getScalarSizeInBits() / 8,
+                     Index, Remainder);
+      if (Remainder != 0)
+        llvm_unreachable("Illegal GEP of a SPIR-V builtin variable");
+
+      IRBuilder<> Builder(Load);
+      Value *Replacement;
+      if (ReplacementFunc->getFunctionType()->getNumParams() == 0) {
+        if (Load->getType() != ScalarTy)
+          llvm_unreachable("Illegal use of a SPIR-V builtin variable");
+        Replacement =
+            setAttrByCalledFunc(Builder.CreateCall(ReplacementFunc, {}));
+      } else {
+        // The function has an index parameter.
+        if (auto *VecTy = dyn_cast<FixedVectorType>(Load->getType())) {
+          // Reconstruct the original global variable vector because
+          // the load type may not match.
+          // global <3 x i64>, load <6 x i32>
+          VecTy = cast<FixedVectorType>(GV->getValueType());
+          if (!Index.isZero() || DL.getTypeSizeInBits(VecTy) !=
+                                     DL.getTypeSizeInBits(Load->getType()))
+            llvm_unreachable("Illegal use of a SPIR-V builtin variable");
+          Replacement = UndefValue::get(VecTy);
+          for (unsigned I = 0; I < VecTy->getNumElements(); I++) {
+            Replacement = Builder.CreateInsertElement(
+                Replacement,
+                setAttrByCalledFunc(
+                    Builder.CreateCall(ReplacementFunc, {Builder.getInt32(I)})),
+                Builder.getInt32(I));
+          }
+          // Insert a bitcast from the reconstructed vector to the load vector
+          // type in case they are different.
+          // Input:
+          // %1 = load <6 x i32>, ptr addrspace(1) %0, align 32
+          // %2 = extractelement <6 x i32> %1, i32 0
+          // %3 = add i32 5, %2
+          // Modified:
+          // < reconstruct global vector elements 0 and 1 >
+          // %2 = insertelement <3 x i64> %0, i64 %1, i32 2
+          // %3 = bitcast <3 x i64> %2 to <6 x i32>
+          // %4 = extractelement <6 x i32> %3, i32 0
+          // %5 = add i32 5, %4
+          Replacement = Builder.CreateBitCast(Replacement, Load->getType());
+        } else if (Load->getType() == ScalarTy) {
+          Replacement = setAttrByCalledFunc(Builder.CreateCall(
+              ReplacementFunc, {Builder.getInt32(Index.getZExtValue())}));
+        } else {
+          llvm_unreachable("Illegal load type of a SPIR-V builtin variable");
+        }
+      }
+      Load->replaceAllUsesWith(Replacement);
+      InstsToRemove.push_back(Load);
+    } else {
+      llvm_unreachable("Illegal use of a SPIR-V builtin variable");
+    }
+  }
+
+  for (Instruction *I : InstsToRemove)
+    I->eraseFromParent();
+}
+
 bool lowerBuiltinVariableToCall(GlobalVariable *GV,
                                 SPIRVBuiltinVariableKind Kind) {
   // There might be dead constant users of GV (for example, SPIRVLowerConstExpr
@@ -1774,7 +2014,7 @@ bool lowerBuiltinVariableToCall(GlobalVariable *GV,
   Module *M = GV->getParent();
   LLVMContext &C = M->getContext();
   std::string FuncName = GV->getName().str();
-  Type *GVTy = GV->getType()->getPointerElementType();
+  Type *GVTy = GV->getValueType();
   Type *ReturnTy = GVTy;
   // Some SPIR-V builtin variables are translated to a function with an index
   // argument.
@@ -1794,111 +2034,11 @@ bool lowerBuiltinVariableToCall(GlobalVariable *GV,
     Func = Function::Create(FT, GlobalValue::ExternalLinkage, MangledName, M);
     Func->setCallingConv(CallingConv::SPIR_FUNC);
     Func->addFnAttr(Attribute::NoUnwind);
-    Func->addFnAttr(Attribute::ReadNone);
     Func->addFnAttr(Attribute::WillReturn);
+    Func->setDoesNotAccessMemory();
   }
 
-  // Collect instructions in these containers to remove them later.
-  std::vector<Instruction *> Loads;
-  std::vector<Instruction *> Casts;
-  std::vector<Instruction *> GEPs;
-
-  auto Replace = [&](std::vector<Value *> Arg, Instruction *I) {
-    auto *Call = CallInst::Create(Func, Arg, "", I);
-    Call->takeName(I);
-    setAttrByCalledFunc(Call);
-    SPIRVDBG(dbgs() << "[lowerBuiltinVariableToCall] " << *I << " -> " << *Call
-                    << '\n';)
-    I->replaceAllUsesWith(Call);
-  };
-
-  // If HasIndexArg is true, we create 3 built-in calls and insertelement to
-  // get 3-element vector filled with ids and replace uses of Load instruction
-  // with this vector.
-  // If HasIndexArg is false, the result of the Load instruction is the value
-  // which should be replaced with the Func.
-  // Returns true if Load was replaced, false otherwise.
-  auto ReplaceIfLoad = [&](User *I) {
-    auto *LD = dyn_cast<LoadInst>(I);
-    if (!LD)
-      return false;
-    std::vector<Value *> Vectors;
-    Loads.push_back(LD);
-    if (HasIndexArg) {
-      auto *VecTy = cast<FixedVectorType>(GVTy);
-      Value *EmptyVec = UndefValue::get(VecTy);
-      Vectors.push_back(EmptyVec);
-      const DebugLoc &DLoc = LD->getDebugLoc();
-      for (unsigned I = 0; I < VecTy->getNumElements(); ++I) {
-        auto *Idx = ConstantInt::get(Type::getInt32Ty(C), I);
-        auto *Call = CallInst::Create(Func, {Idx}, "", LD);
-        if (DLoc)
-          Call->setDebugLoc(DLoc);
-        setAttrByCalledFunc(Call);
-        auto *Insert = InsertElementInst::Create(Vectors.back(), Call, Idx);
-        if (DLoc)
-          Insert->setDebugLoc(DLoc);
-        Insert->insertAfter(Call);
-        Vectors.push_back(Insert);
-      }
-
-      Value *Ptr = LD->getPointerOperand();
-
-      if (isa<FixedVectorType>(Ptr->getType()->getPointerElementType())) {
-        LD->replaceAllUsesWith(Vectors.back());
-      } else {
-        auto *GEP = dyn_cast<GetElementPtrInst>(Ptr);
-        assert(GEP && "Unexpected pattern!");
-        assert(GEP->getNumIndices() == 2 && "Unexpected pattern!");
-        Value *Idx = GEP->getOperand(2);
-        Value *Vec = Vectors.back();
-        auto *NewExtract = ExtractElementInst::Create(Vec, Idx);
-        NewExtract->insertAfter(cast<Instruction>(Vec));
-        LD->replaceAllUsesWith(NewExtract);
-      }
-
-    } else {
-      Replace({}, LD);
-    }
-
-    return true;
-  };
-
-  // Go over the GV users, find Load and ExtractElement instructions and
-  // replace them with the corresponding function call.
-  for (auto *UI : GV->users()) {
-    // There might or might not be an addrspacecast instruction.
-    if (auto *ASCast = dyn_cast<AddrSpaceCastInst>(UI)) {
-      Casts.push_back(ASCast);
-      for (auto *CastUser : ASCast->users()) {
-        if (ReplaceIfLoad(CastUser))
-          continue;
-        if (auto *GEP = dyn_cast<GetElementPtrInst>(CastUser)) {
-          GEPs.push_back(GEP);
-          for (auto *GEPUser : GEP->users()) {
-            if (!ReplaceIfLoad(GEPUser))
-              llvm_unreachable("Unexpected pattern!");
-          }
-        } else {
-          llvm_unreachable("Unexpected pattern!");
-        }
-      }
-    } else if (!ReplaceIfLoad(UI)) {
-      llvm_unreachable("Unexpected pattern!");
-    }
-  }
-
-  auto Erase = [](std::vector<Instruction *> &ToErase) {
-    for (Instruction *I : ToErase) {
-      assert(I->hasNUses(0));
-      I->eraseFromParent();
-    }
-  };
-  // Order of erasing is important.
-  Erase(Loads);
-  Erase(GEPs);
-  Erase(Casts);
-
+  replaceUsesOfBuiltinVar(GV, APInt(64, 0), Func, GV);
   return true;
 }
 
@@ -1927,11 +2067,23 @@ bool postProcessBuiltinReturningStruct(Function *F) {
   SmallVector<Instruction *, 32> InstToRemove;
   for (auto *U : F->users()) {
     if (auto *CI = dyn_cast<CallInst>(U)) {
-      auto *ST = cast<StoreInst>(*(CI->user_begin()));
+      IRBuilder<> Builder(CI->getParent());
+      Builder.SetInsertPoint(CI);
+      SmallVector<User *, 5> Users(CI->users());
+      Value *A = nullptr;
+      for (auto *U : Users) {
+        if (auto *SI = dyn_cast<StoreInst>(U)) {
+          A = SI->getPointerOperand();
+          InstToRemove.push_back(SI);
+          break;
+        }
+      }
+      if (!A) {
+        A = Builder.CreateAlloca(F->getReturnType());
+      }
       std::vector<Type *> ArgTys;
       getFunctionTypeParameterTypes(F->getFunctionType(), ArgTys);
-      ArgTys.insert(ArgTys.begin(),
-                    PointerType::get(F->getReturnType(), SPIRAS_Private));
+      ArgTys.insert(ArgTys.begin(), A->getType());
       auto *NewF =
           getOrCreateFunction(M, Type::getVoidTy(*Context), ArgTys, Name);
       NewF->addParamAttr(0, Attribute::get(*Context,
@@ -1939,10 +2091,13 @@ bool postProcessBuiltinReturningStruct(Function *F) {
                                            F->getReturnType()));
       NewF->setCallingConv(F->getCallingConv());
       auto Args = getArguments(CI);
-      Args.insert(Args.begin(), ST->getPointerOperand());
-      auto *NewCI = CallInst::Create(NewF, Args, CI->getName(), CI);
+      Args.insert(Args.begin(), A);
+      CallInst *NewCI = Builder.CreateCall(NewF, Args, CI->getName());
       NewCI->setCallingConv(CI->getCallingConv());
-      InstToRemove.push_back(ST);
+      SmallVector<User *, 5> CIUsers(CI->users());
+      for (auto *CIUser : CIUsers) {
+        CIUser->replaceUsesOfWith(CI, A);
+      }
       InstToRemove.push_back(CI);
     }
   }
@@ -2019,12 +2174,13 @@ bool postProcessBuiltinsWithArrayArguments(Module *M, bool IsCpp) {
 namespace {
 class SPIRVFriendlyIRMangleInfo : public BuiltinFuncMangleInfo {
 public:
-  SPIRVFriendlyIRMangleInfo(spv::Op OC, ArrayRef<Type *> ArgTys)
-      : OC(OC), ArgTys(ArgTys) {}
+  SPIRVFriendlyIRMangleInfo(spv::Op OC, ArrayRef<Type *> ArgTys,
+                            ArrayRef<SPIRVValue *> Ops)
+      : OC(OC), ArgTys(ArgTys), Ops(Ops) {}
 
   void init(StringRef UniqUnmangledName) override {
     UnmangledName = UniqUnmangledName.str();
-    switch (OC) {
+    switch (static_cast<unsigned>(OC)) {
     case OpConvertUToF:
     case OpUConvert:
     case OpSatConvertUToS:
@@ -2065,11 +2221,16 @@ public:
     case OpGroupNonUniformShuffleDown:
       addUnsignedArg(2);
       break;
+    case OpGroupNonUniformRotateKHR:
+      if (ArgTys.size() == 4)
+        addUnsignedArg(3);
+      break;
     case OpGroupNonUniformInverseBallot:
     case OpGroupNonUniformBallotFindLSB:
     case OpGroupNonUniformBallotFindMSB:
       addUnsignedArg(1);
       break;
+    case OpBitFieldSExtract:
     case OpGroupNonUniformBallotBitExtract:
       addUnsignedArg(1);
       addUnsignedArg(2);
@@ -2090,6 +2251,7 @@ public:
     case OpGroupNonUniformLogicalXor:
       addUnsignedArg(3);
       break;
+    case OpBitFieldInsert:
     case OpGroupNonUniformUMax:
     case OpGroupNonUniformUMin:
       addUnsignedArg(2);
@@ -2140,6 +2302,7 @@ public:
     case OpSubgroupAvcImeSetSingleReferenceINTEL:
       addUnsignedArg(1);
       break;
+    case OpBitFieldUExtract:
     case OpSubgroupAvcImeInitializeINTEL:
     case OpSubgroupAvcMceSetMotionVectorCostFunctionINTEL:
     case OpSubgroupAvcSicSetIntraLumaModeCostFunctionINTEL:
@@ -2165,6 +2328,28 @@ public:
     case OpSubgroupAvcSicConfigureSkcINTEL:
       addUnsignedArgs(0, 4);
       break;
+    case OpUDotKHR:
+    case OpUDotAccSatKHR:
+      addUnsignedArg(-1);
+      break;
+    case OpSUDotKHR:
+    case OpSUDotAccSatKHR:
+      addUnsignedArg(1);
+      break;
+    case OpImageWrite: {
+      size_t Idx = getImageOperandsIndex(OC);
+      if (Ops.size() > Idx) {
+        auto ImOp = static_cast<SPIRVConstant *>(Ops[Idx])->getZExtIntValue();
+        if (ImOp & ImageOperandsMask::ImageOperandsZeroExtendMask)
+          addUnsignedArg(2);
+      }
+      break;
+    }
+    case internal::OpConvertHandleToImageINTEL:
+    case internal::OpConvertHandleToSamplerINTEL:
+    case internal::OpConvertHandleToSampledImageINTEL:
+      addUnsignedArg(0);
+      break;
     default:;
       // No special handling is needed
     }
@@ -2173,6 +2358,7 @@ public:
 private:
   spv::Op OC;
   ArrayRef<Type *> ArgTys;
+  ArrayRef<SPIRVValue *> Ops;
 };
 class OpenCLStdToSPIRVFriendlyIRMangleInfo : public BuiltinFuncMangleInfo {
 public:
@@ -2242,9 +2428,9 @@ std::string getSPIRVFriendlyIRFunctionName(OCLExtOpKind ExtOpId,
 }
 
 std::string getSPIRVFriendlyIRFunctionName(const std::string &UniqName,
-                                           spv::Op OC,
-                                           ArrayRef<Type *> ArgTys) {
-  SPIRVFriendlyIRMangleInfo MangleInfo(OC, ArgTys);
+                                           spv::Op OC, ArrayRef<Type *> ArgTys,
+                                           ArrayRef<SPIRVValue *> Ops) {
+  SPIRVFriendlyIRMangleInfo MangleInfo(OC, ArgTys, Ops);
   return mangleBuiltin(UniqName, ArgTys, &MangleInfo);
 }
 
